# Copyright 2020 - 2022 MONAI Consortium
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#     http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import scipy.ndimage as ndimage
import torch


def resample_3d(img, target_size):
    imx, imy, imz = img.shape
    tx, ty, tz = target_size
    zoom_ratio = (float(tx) / float(imx), float(ty) / float(imy), float(tz) / float(imz))
    img_resampled = ndimage.zoom(img, zoom_ratio, order=0, prefilter=False)
    return img_resampled


def dice(x, y): #output,label
    intersect = np.sum(np.sum(np.sum(x * y)))
    y_sum = np.sum(np.sum(np.sum(y)))
    x_sum = np.sum(np.sum(np.sum(x)))
    if x_sum == 0 and y_sum==0:
        return None,'TN'
    if x_sum == 0 and y_sum != 0:
        return 0.0,'FN'
    if y_sum == 0:
        return 2 * intersect / (x_sum + y_sum),'FP'
    if y_sum != 0:
        return 2 * intersect / (x_sum + y_sum),'P'


class AverageMeter(object):
    def __init__(self):
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n=1):
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = np.where(self.count > 0, self.sum / self.count, self.sum)


def distributed_all_gather(
    tensor_list, valid_batch_size=None, out_numpy=False, out_value=False, out_list=False, world_size=None, no_barrier=False, is_valid=None
):
    if world_size is None:
        world_size = torch.distributed.get_world_size()
    if valid_batch_size is not None:
        valid_batch_size = min(valid_batch_size, world_size)
    elif is_valid is not None:
        is_valid = torch.tensor(bool(is_valid), dtype=torch.bool, device=tensor_list[0].device)
    if not no_barrier:
        torch.distributed.barrier()
    tensor_list_out = []
    with torch.no_grad():
        if is_valid is not None:
            is_valid_list = [torch.zeros_like(is_valid) for _ in range(world_size)]
            torch.distributed.all_gather(is_valid_list, is_valid)
            is_valid = [x.item() for x in is_valid_list]
        for tensor in tensor_list:
            gather_list = [torch.zeros_like(tensor) for _ in range(world_size)]
            torch.distributed.all_gather(gather_list, tensor)
            if valid_batch_size is not None:
                gather_list = gather_list[:valid_batch_size]
            elif is_valid is not None:
                gather_list = [g for g, v in zip(gather_list, is_valid_list) if v]
            if out_numpy:
                gather_list = [t.cpu().numpy() for t in gather_list]
            if out_list:
                gather_list = [t.cpu().numpy().tolist() for t in gather_list]
            if out_value:
                gather_list = [t.cpu().value() for t in gather_list]
            tensor_list_out.append(gather_list)
    return tensor_list_out



class LR_Scheduler(object):
    def __init__(self, base_lr, num_epochs, warmup=100, mode='poly'):
        self.mode = mode
        self.lr = base_lr
        self.num_epochs = num_epochs
        self.warmup = warmup

    def __call__(self, optimizer, epoch):
        if self.mode == 'poly':
            now_lr = round(self.lr * np.power(1 - np.float32(epoch)/np.float32(self.num_epochs), 0.9), 8)
        elif self.mode == 'warmup':
            if epoch < self.warmup*2:
                now_lr = round(0.5 * self.lr * (1.0 + math.cos(((np.float32(epoch)/np.float32(self.warmup))*math.pi))),8)
            else:
                now_lr = round(self.lr * np.power(1 - (np.float32(epoch) - np.float32(self.warmup*2))/(np.float32(self.num_epochs)-np.float32(self.warmup*2)), 0.9), 8)
        elif self.mode == 'cousinewarmup':
            if self.warmup == 0:
                if epoch < 100:
                    now_lr = round(self.lr * (math.sin(((np.float32(epoch))/(np.float32(100.0 * 2.0)))*math.pi)),8)
                else:
                    now_lr = round(0.5 * self.lr * (1.0 + math.cos(((np.float32(epoch) - np.float32(100.0))/(np.float32(self.num_epochs)-np.float32(100.0)))*math.pi)), 8)
            else:
                if epoch < self.warmup*2:
                    now_lr = round(0.5 * self.lr * (1.0 + math.cos(((np.float32(epoch)/np.float32(self.warmup))*math.pi))),8)
                else:
                    now_lr = round(0.5 * self.lr * (1.0 + math.cos(((np.float32(epoch) - np.float32(self.warmup*2))/(np.float32(self.num_epochs)-np.float32(self.warmup*2)))*math.pi)), 8)
        elif self.mode == 'warmuppoly':
            if epoch < 50:
                now_lr = round(self.lr * (((np.float32(epoch+1.0))/(np.float32(50.0)))),8)
            else:
                now_lr = round(self.lr * np.power(1 - (np.float32(epoch) - np.float32(50.0))/(np.float32(self.num_epochs+1)-np.float32(50.0)), 0.9), 8)
        self._adjust_learning_rate(optimizer, now_lr)
        return now_lr

    def _adjust_learning_rate(self, optimizer, lr):
        optimizer.param_groups[0]['lr'] = lr